/*
**  CXSC is a C++ library for eXtended Scientific Computing (V 2.1)
**
**  Copyright (C) 1990-2000 Institut fuer Angewandte Mathematik,
**                          Universitaet Karlsruhe, Germany
**            (C) 2000-2005 Wiss. Rechnen/Softwaretechnologie
**                          Universitaet Wuppertal, Germany   
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* CVS $Id: l_rmath.inl,v 1.13 2005/09/01 12:56:40 cxsc Exp $ */

#include "l_interv.hpp"
#include "l_imath.hpp"

namespace cxsc {

inline l_real sqr(const l_real &x) throw() { return (x*x); }
inline l_real sqrt(const l_real & x,int n) { return mid(sqrt(_l_interval(x),n)); }

inline l_real sqrtx2y2(const l_real& x, 
                       const l_real& y) throw() // Blomquist 6.12.02
// Calculation of an approximation of sqrt(x^2+y^2).
// In general the maximum precision is stagprec=19, predifined by the used
// sqrt-function declared in l_rmath.hpp.
// If the difference |exa-exb| of the exponents (base 2) is sufficient high,
// precision and accuracy can be choosen greater 19.
{ 
    int stagsave = stagprec, stagcalc, stagmax = 19;
    l_real a,b,r,r1;
    int exa,exb,ex;
    a = x;  b = y;
    exa = expo(a[1]);
    exb = expo(b[1]);
    if (exb > exa)
    {  // Permutation of a,b:
	r = a;  a = b;  b = r;
	ex = exa;  exa = exb;  exb = ex;
    } // |a| >= |b|
    if (sign(a[1]) < 0) a = -a;  // a == abs(a);
    if (sign(b[1]==0)) return a;

    // |a| = a >= |b| > 0:
    ex = 0;  // initialization
    if (6*exb < 5*exa-1071)
    {   // approximation with a + 0.5*b^2/a - b^4/(8*a^3), if the next
        // Taylor term b^6/(16*a^5) is less than 2^-1074==minreal;
        // minreal is the smallest positive number in IEEE-format.
	r1 = (b/a);
	r = r1*r1;  
	times2pown(r,-2);   // r = r/4;
	r = 1 - r;
	r1 *= b;
	times2pown(r1,-1);  // r = r/2;
	r *= r1;
	r += a; // r = a + 0.5*b^2/a - b^4/(8*a^3);
    } else
    {   // Scaling ubwards or downwards to achiev optimal accuracy and 
        // to avoid overflow by accu-reading:
        stagcalc = stagprec;  
	if (stagcalc > stagmax) stagcalc = stagmax; // using sqrt(...) a 
	stagprec = stagcalc; // higher precision than stagmax makes no sense!

	ex = 511 - exa; // scaling factor to avoide overflow by accu-reading
	if (ex < 0)
	{   // sqrt(a^2+b^2) = a*sqrt( 1+(b/a)^2 )
	    r = b/a;
	    r = a*sqrt(1+r*r);  // sqrt(...) declared in l_rmath.hpp
	} else
	{   
            times2pown(a,ex);   // exact scaling with ex >= 0
	    times2pown(b,ex);   // exact scaling eith ex >= 0
	    dotakku[0] = 0;
	    accumulate(dotakku[0],a,a);
	    accumulate(dotakku[0],b,b);
	    r = dotakku[0]; // r with no overflow!
	    r = sqrt(r);  // sqrt(...) declared in l_rmath.hpp
	    times2pown(r,-ex);  // back-scaling 
	}
	stagprec = stagsave;  // restore old stagprec
    }
    return r;
} // sqrtx2y2(...)

inline l_real sqrt1px2(const l_real& x) throw()
// Inclusion of sqrt(1+x^2); Blomquist, 13.12.02;
// With stagmax=19 we get about 16*19=304 exact decimal digits.
{
    l_real y,t;
    int stagsave, stagmax=19;
    stagsave = stagprec;
    if (stagprec > stagmax) stagprec = stagmax;
    if (expo(x[1]) > 260)
    {  // sqrt(1+x^2) = |x| + 1/(2*|x|)
	y = abs(x);
	t = 1/y;
	times2pown(t,-1);
	y += t;
    } else y = sqrt(1+x*x);
    stagprec = stagsave;
    y = adjust(y);
    return y;
}

inline l_real sqrtp1m1(const l_real & x) throw() // Blomquist, 05.08.03
     { return mid(sqrtp1m1(_l_interval(x))); }
inline l_real sqrtx2m1(const l_real & x) throw() 
     { return mid(sqrtx2m1(_l_interval(x))); } // sqrt(x^2-1);  Blomquist, 13.04.04;
inline l_real sqrt1mx2(const l_real& x) throw() 
     { return mid(sqrt1mx2(_l_interval(x)));}
inline l_real ln_sqrtx2y2(const l_real& x, const l_real& y) throw()
     { return mid(ln_sqrtx2y2(_l_interval(x),_l_interval(y)));}
inline l_real acoshp1(const l_real& x)
     { return mid(acoshp1(l_interval(x)));}


inline l_real sin(const l_real & x) throw() { return mid(sin(_l_interval(x))); }
inline l_real cos(const l_real & x) throw() { return mid(cos(_l_interval(x))); }
inline l_real tan(const l_real & x) throw() { return mid(tan(_l_interval(x))); }
inline l_real cot(const l_real & x) throw() { return mid(cot(_l_interval(x))); }

inline l_real asin(const l_real & x)         { return mid(asin(_l_interval(x))); }
inline l_real acos(const l_real & x)         { return mid(acos(_l_interval(x))); }
inline l_real atan(const l_real & x)         { return mid(atan(_l_interval(x))); }
inline l_real acot(const l_real & x)         { return mid(acot(_l_interval(x))); }

inline l_real exp(const l_real & x) throw() { return mid(exp(_l_interval(x))); }
inline l_real expmx2(const l_real& x) throw() { return mid(expmx2(l_interval(x)));}


inline l_real ln(const l_real & x)          { return mid(ln(_l_interval(x))); }

inline l_real sinh(const l_real & x) throw() { return mid(sinh(_l_interval(x))); }
inline l_real cosh(const l_real & x) throw() { return mid(cosh(_l_interval(x))); }
inline l_real tanh(const l_real & x) throw() { return mid(tanh(_l_interval(x))); }
inline l_real coth(const l_real & x) throw() { return mid(coth(_l_interval(x))); }

inline l_real asinh(const l_real & x)         { return mid(asinh(_l_interval(x))); }
inline l_real acosh(const l_real & x)         { return mid(acosh(_l_interval(x))); }
inline l_real atanh(const l_real & x)         { return mid(atanh(_l_interval(x))); }
inline l_real acoth(const l_real & x)         { return mid(acoth(_l_interval(x))); }

inline l_real pow(const l_real & x,const l_real & expo) { return mid(pow(_l_interval(x),_l_interval(expo))); }

} // namespace cxsc



